@@ -2,23 +2,36 @@
 
 import json
 import logging
-import pydantic
+from pathlib import Path
 import re
-import requests
 import socket
 import sys
-from typing import (List, Optional, Tuple, Union)
+from typing import (Dict, List, Optional, Tuple, Union)
 import urllib3
 from urllib.parse import quote
 
+import pydantic
+from pydantic.main import ModelMetaclass
+import requests
+
 from trs_cli.errors import (
     exception_handler,
     ContentTypeUnavailable,
+    FileInformationUnavailable,
     InvalidURI,
+    InvalidPayload,
     InvalidResourceIdentifier,
     InvalidResponseError,
 )
-from trs_cli.models import (Error, FileWrapper, ToolFile)
+from trs_cli.models import (
+    Error,
+    FileType,
+    FileWrapper,
+    ToolFile,
+    Tool,
+    ToolClass,
+    ToolClassRegister,
+)
 
 logger = logging.getLogger(__name__)
 sys.excepthook = exception_handler
@@ -88,6 +101,121 @@ class TRSClient():
         self.headers = {}
         logger.info(f"Instantiated client for: {self.uri}")
 
+    def post_tool_class(
+        self,
+        payload: Dict,
+        accept: str = 'application/json',
+        token: Optional[str] = None,
+    ) -> str:
+        """Register a tool class.
+
+        Arguments:
+            payload: Tool class data.
+            accept: Requested content type.
+            token: Bearer token for authentication. Set if required by TRS
+                implementation and if not provided when instatiating client or
+                if expired.
+
+        Returns:
+            ID of registered TRS toolClass in case of a `200` response, or an
+            instance of `Error` for all other responses.
+        Raises:
+            requests.exceptions.ConnectionError: A connection to the provided
+                TRS instance could not be established.
+            trs_cli.errors.InvalidPayload: The object data payload could not
+                be validated against the API schema.
+            trs_cli.errors.InvalidResponseError: The response could not be
+                validated against the API schema.
+        """
+        # validate requested content type and get request headers
+        self._validate_content_type(
+            requested_type=accept,
+            available_types=['application/json'],
+        )
+        self._get_headers(
+            content_accept=accept,
+            content_type='application/json',
+            token=token,
+        )
+
+        # build request URL
+        url = f"{self.uri}/toolClasses"
+        logger.info(f"Connecting to '{url}'...")
+
+        # validate payload
+        try:
+            ToolClassRegister(**payload).dict()
+        except pydantic.ValidationError:
+            raise InvalidPayload(
+                "The tool class data could not be validated against API "
+                "schema."
+            )
+
+        # send request
+        response = self._send_request_and_validate_response(
+            url=url,
+            method='post',
+            payload=payload,
+        )
+        logger.info(
+            "Registered tool class"
+        )
+        return response  # type: ignore
+
+    def get_tool(
+        self,
+        id: str,
+        accept: str = 'application/json',
+        token: Optional[str] = None,
+    ) -> Union[Tool, Error]:
+        """Retrieve tool with the specified identifier.
+
+        Arguments:
+            tool_id: Implementation-specific TRS identifier hostname-based
+               TRS URI pointing to a given tool
+            accept: Requested content type.
+            token: Bearer token for authentication. Set if required by TRS
+                implementation and if not provided when instatiating client or
+                if expired.
+
+        Returns:
+            Unmarshalled TRS response as either an instance of `Tool`
+            in case of a `200` response, or an instance of `Error` for all
+            other JSON reponses.
+
+        Raises:
+            requests.exceptions.ConnectionError: A connection to the provided
+                TRS instance could not be established.
+            trs_cli.errors.InvalidResponseError: The response could not be
+                validated against the API schema.
+        """
+        # validate requested content type and get request headers
+        self._validate_content_type(
+            requested_type=accept,
+            available_types=['application/json', 'text/plain'],
+        )
+        self._get_headers(
+            content_accept=accept,
+            token=token,
+        )
+
+        # get/sanitize tool identifier
+        _id, _ = self._get_tool_id_version_id(tool_id=id)
+
+        # build request URL
+        url = f"{self.uri}/tools/{_id}"
+        logger.info(f"Connecting to '{url}'...")
+
+        # send request
+        response = self._send_request_and_validate_response(
+            url=url,
+            validation_class_200=Tool,
+        )
+        logger.info(
+            "Retrieved tool"
+        )
+        return response  # type: ignore
+
     def get_descriptor(
         self,
         type: str,
@@ -96,7 +224,8 @@ class TRSClient():
         accept: str = 'application/json',
         token: Optional[str] = None
     ) -> Union[FileWrapper, Error]:
-        """Get the tool descriptor for the specified tool.
+        """Retrieve the file wrapper for the primary descriptor of a specified
+        tool version and descriptor type.
 
         Arguments:
             type: The output type of the descriptor. Plain types return
@@ -113,6 +242,7 @@ class TRSClient():
                 registry. It is optional if a TRS URI is passed and includes
                 version information. If provided nevertheless, then the
                 `version_id` retrieved from the TRS URI is overridden.
+            accept: Requested content type.
             token: Bearer token for authentication. Set if required by TRS
                 implementation and if not provided when instatiating client or
                 if expired.
@@ -128,14 +258,15 @@ class TRSClient():
             trs_cli.errors.InvalidResponseError: The response could not be
                 validated against the API schema.
         """
-        # validate requested content type, set token and get request headers
+        # validate requested content type and get request headers
         self._validate_content_type(
             requested_type=accept,
             available_types=['application/json', 'text/plain'],
         )
-        if token:
-            self.token = token
-        self._get_headers(content_accept=accept)
+        self._get_headers(
+            content_accept=accept,
+            token=token,
+        )
 
         # get/sanitize tool and version identifiers
         _id, _version_id = self._get_tool_id_version_id(
@@ -150,44 +281,15 @@ class TRSClient():
         )
         logger.info(f"Connecting to '{url}'...")
 
-        # send request and handle exceptions and error responses
-        try:
-            response = requests.get(
-                url=url,
-                headers=self.headers,
-            )
-        except (
-            requests.exceptions.ConnectionError,
-            socket.gaierror,
-            urllib3.exceptions.NewConnectionError,
-        ):
-            raise requests.exceptions.ConnectionError(
-                "Could not connect to API endpoint."
-            )
-        if not response.status_code == 200:
-            try:
-                response_val = Error(**response.json())
-            except (
-                json.decoder.JSONDecodeError,
-                pydantic.ValidationError,
-            ):
-                raise InvalidResponseError(
-                    "Response could not be validated against API schema."
-                )
-            logger.warning("Received error response.")
-        else:
-            try:
-                response_val = FileWrapper(**response.json())
-            except pydantic.ValidationError:
-                raise InvalidResponseError(
-                    "Response could not be validated against API schema."
-                )
-            logger.info(
-                f"Retrieved descriptor of type '{type}' "
-                f"for tool '{_id}', version '{_version_id}'."
-            )
-
-        return response_val
+        # send request
+        response = self._send_request_and_validate_response(
+            url=url,
+            validation_class_200=FileWrapper,
+        )
+        logger.info(
+            "Retrieved descriptor"
+        )
+        return response  # type: ignore
 
     def get_descriptor_by_path(
         self,
@@ -199,7 +301,8 @@ class TRSClient():
         accept: str = 'application/json',
         token: Optional[str] = None
     ) -> Union[FileWrapper, Error]:
-        """Get the tool descriptor for the specified tool.
+        """Retrieve the file wrapper for an indicated file for the specified
+        tool version and descriptor type.
 
         Arguments:
             type: The output type of the descriptor. Plain types return
@@ -219,6 +322,7 @@ class TRSClient():
                 version information. If provided nevertheless, then the
                 `version_id` retrieved from the TRS URI is overridden.
             is_encoded: Value of `path` is already percent/URL-encoded.
+            accept: Requested content type.
             token: Bearer token for authentication. Set if required by TRS
                 implementation and if not provided when instatiating client or
                 if expired.
@@ -234,14 +338,15 @@ class TRSClient():
             trs_cli.errors.InvalidResponseError: The response could not be
                 validated against the API schema.
         """
-        # validate requested content type, set token and get request headers
+        # validate requested content type and get request headers
         self._validate_content_type(
             requested_type=accept,
             available_types=['application/json', 'text/plain'],
         )
-        if token:
-            self.token = token
-        self._get_headers(content_accept=accept)
+        self._get_headers(
+            content_accept=accept,
+            token=token,
+        )
 
         # get/sanitize tool and version identifiers
         _id, _version_id = self._get_tool_id_version_id(
@@ -257,44 +362,15 @@ class TRSClient():
         )
         logger.info(f"Connecting to '{url}'...")
 
-        # send request and handle exceptions and error responses
-        try:
-            response = requests.get(
-                url=url,
-                headers=self.headers,
-            )
-        except (
-            requests.exceptions.ConnectionError,
-            socket.gaierror,
-            urllib3.exceptions.NewConnectionError,
-        ):
-            raise requests.exceptions.ConnectionError(
-                "Could not connect to API endpoint."
-            )
-        if not response.status_code == 200:
-            try:
-                response_val = Error(**response.json())
-            except (
-                json.decoder.JSONDecodeError,
-                pydantic.ValidationError,
-            ):
-                raise InvalidResponseError(
-                    "Response could not be validated against API schema."
-                )
-            logger.warning("Received error response.")
-        else:
-            try:
-                response_val = FileWrapper(**response.json())
-            except pydantic.ValidationError:
-                raise InvalidResponseError(
-                    "Response could not be validated against API schema."
-                )
-            logger.info(
-                f"Retrieved file associated with descriptor type '{type}' "
-                f"for tool '{_id}', version '{_version_id}'."
-            )
-
-        return response_val
+        # send request
+        response = self._send_request_and_validate_response(
+            url=url,
+            validation_class_200=FileWrapper,
+        )
+        logger.info(
+            "Retrieved descriptor"
+        )
+        return response  # type: ignore
 
     def get_files(
         self,
@@ -304,7 +380,9 @@ class TRSClient():
         format: Optional[str] = None,
         token: Optional[str] = None
     ) -> Union[List[ToolFile], Error]:
-        """Get the tool files for the specified tool.
+        """Retrieve file information for the specified tool version and
+        descriptor type.
+
         Arguments:
             type: The output type of the descriptor. Plain types return
                 the bare descriptor while the "non-plain" types return a
@@ -323,32 +401,34 @@ class TRSClient():
             token: Bearer token for authentication. Set if required by TRS
                 implementation and if not provided when instatiating client or
                 if expired.
+
         Returns:
             Unmarshalled TRS response as either an instance of `ToolFile` in
             case of a `200` response, or an instance of `Error` for all other
             JSON reponses.
+
         Raises:
             requests.exceptions.ConnectionError: A connection to the provided
                 TRS instance could not be established.
             trs_cli.errors.InvalidResponseError: The response could not be
                 validated against the API schema.
         """
-        # validate requested content type, set token and get request headers
+        # validate requested content type and get request headers
         if format is None:
             query_format = ""
-            content_accept = 'application/json'
+            accept = 'application/json'
         elif format == 'zip':
             query_format = "?format=zip"
-            content_accept = 'application/zip'
+            accept = 'application/zip'
         else:
-            logger.error(
-                "Only 'zip' is an allowed value for the format parameter. "
-                "Omit query parameter to request a JSON response instead."
+            raise ContentTypeUnavailable(
+                "Only 'zip' is allowed for parameter 'format'; omit query"
+                "parameter to request JSON instead"
             )
-            raise ContentTypeUnavailable
-        if token:
-            self.token = token
-        self._get_headers(content_accept=content_accept)
+        self._get_headers(
+            content_accept=accept,
+            token=token,
+        )
 
         # get/sanitize tool and version identifiers
         _id, _version_id = self._get_tool_id_version_id(
@@ -363,46 +443,155 @@ class TRSClient():
         )
         logger.info(f"Connecting to '{url}'...")
 
-        # send request and handle exceptions and error responses
+        # send request
+        response = self._send_request_and_validate_response(
+            url=url,
+            validation_class_200=(ToolFile, ),
+        )
+        logger.info(
+            "Retrieved files"
+        )
+        return response  # type: ignore
+
+    def get_tool_classes(
+        self,
+        accept: str = 'application/json',
+        token: Optional[str] = None
+    ) -> Union[List[ToolClass], Error]:
+        """Retrieve tool classes.
+
+        Arguments:
+            accept: Requested content type.
+            token: Bearer token for authentication. Set if required by TRS
+                implementation and if not provided when instatiating client or
+                if expired.
+
+        Returns:
+            Unmarshalled TRS response as either a list of `ToolClass` objects
+            in case of a `200` response, or an instance of `Error` for all
+            other JSON reponses.
+
+        Raises:
+            requests.exceptions.ConnectionError: A connection to the provided
+                TRS instance could not be established.
+            trs_cli.errors.InvalidResponseError: The response could not be
+                validated against the API schema.
+        """
+        # validate requested content type and get request headers
+        self._validate_content_type(
+            requested_type=accept,
+            available_types=['application/json', 'text/plain'],
+        )
+        self._get_headers(
+            content_accept=accept,
+            token=token,
+        )
+
+        # build request URL
+        url = f"{self.uri}/toolClasses"
+        logger.info(f"Connecting to '{url}'...")
+
+        # send request
+        response = self._send_request_and_validate_response(
+            url=url,
+            validation_class_200=(ToolClass, ),
+        )
+        logger.info(
+            "Retrieved tool classes"
+        )
+        return response  # type: ignore
+
+    def retrieve_files(
+        self,
+        out_dir: Union[str, Path],
+        type: str,
+        id: str,
+        version_id: Optional[str] = None,
+        is_encoded: bool = False,
+        token: Optional[str] = None,
+    ) -> Dict[str, List[str]]:
+        """Write tool version file contents for a given descriptor type to
+        files.
+
+        Arguments:
+            out_dir: Directory to write requested files to. Will be attempted
+                to create if it does not exist.
+            type: The output type of the descriptor. Plain types return
+                the bare descriptor while the "non-plain" types return a
+                descriptor wrapped with metadata. Allowable values include
+                "CWL", "WDL", "NFL", "GALAXY", "PLAIN_CWL", "PLAIN_WDL",
+                "PLAIN_NFL", "PLAIN_GALAXY".
+            id: A unique identifier of the tool, scoped to this registry OR
+                a hostname-based TRS URI. If TRS URIs include the version
+                information, passing a `version_id` is optional.
+            version_id: An optional identifier of the tool version, scoped
+                to this registry. It is optional if version info is included
+                in the TRS URI. If passed, then the existing `version_id`
+                retreived from the TRS URI is overridden.
+            is_encoded: Values or relative paths of files are already
+                percent/URL-encoded.
+
+        Returns:
+            Dictionary of `FileType` enumerator values (e.g., `TEST_FILE`,
+            `PRIMARY_DESCRIPTOR`) as keys and a list of paths, relative to
+            `out_dir` as values.
+        """
+        # if not exists, try to create output directory
+        out_dir = Path(out_dir)
         try:
-            response = requests.get(
-                url=url,
-                headers=self.headers,
-            )
-        except (
-            requests.exceptions.ConnectionError,
-            socket.gaierror,
-            urllib3.exceptions.NewConnectionError,
-        ):
-            raise requests.exceptions.ConnectionError(
-                "Could not connect to API endpoint."
-            )
-        if not response.status_code == 200:
-            try:
-                response_val = Error(**response.json())
-            except (
-                json.decoder.JSONDecodeError,
-                pydantic.ValidationError,
-            ):
-                raise InvalidResponseError(
-                    "Response could not be validated against API schema."
-                )
-            logger.warning("Received error response.")
-        else:
-            try:
-                response_val = [
-                    ToolFile(**tool) for tool in response.json()
-                ]
-            except pydantic.ValidationError:
-                raise InvalidResponseError(
-                    "Response could not be validated against API schema."
-                )
-            logger.info(
-                f"Retrieved files of type '{type}' for tool '{id}',"
-                f"version '{version_id}'."
+            Path(out_dir).mkdir(parents=True, exist_ok=True)
+        except OSError:
+            raise OSError("Could not create output directory")
+
+        # get file information
+        files = self.get_files(
+            type=type,
+            id=id,
+            version_id=version_id,
+            token=token,
+        )
+        if isinstance(files, Error):
+            raise FileInformationUnavailable(
+                "File information unavailable"
             )
 
-        return response_val
+        # get path of primary descriptor
+        paths_by_type = {item.value: [] for item in FileType}
+        for _file in files:
+            if _file.path is not None:
+                paths_by_type[_file.file_type.value].append(_file.path)
+
+        # get file wrappers
+        file_wrappers = {}
+        for _f in files:
+            if not hasattr(_f, 'path') or _f.path is None:
+                raise FileInformationUnavailable(
+                    f"Path information unavailable for file object: {_f}"
+                )
+            file_wrapper = self.get_descriptor_by_path(
+                type=type,
+                path=_f.path,
+                id=id,
+                version_id=version_id,
+                is_encoded=is_encoded,
+                token=token,
+            )
+            if not isinstance(file_wrapper, FileWrapper):
+                raise FileInformationUnavailable(
+                    "Content unavailable for file at path '{_f.path}'"
+                )
+            file_wrappers[_f.path] = file_wrapper.content
+
+        # write contents to files
+        for path, content in file_wrappers.items():
+            out_path = out_dir / path
+            try:
+                with open(out_path, "w") as _fp:
+                    _fp.write(content)
+            except OSError:
+                raise OSError(f"Could not write file '{str(out_path)}'")
+
+        return paths_by_type
 
     def _get_host(
         self,
@@ -475,22 +664,25 @@ class TRSClient():
         ret_version_id: Optional[str] = None
 
         if tool_id is None and version_id is None:
-            logger.error("No TRS URI, tool or version identifier supplied.")
-            raise InvalidResourceIdentifier
+            raise InvalidResourceIdentifier(
+                "No TRS URI, tool or version identifier supplied"
+            )
 
         if tool_id is not None:
             match = re.search(self._RE_TRS_URI_OR_TOOL_ID, tool_id, re.I)
             if match is None:
-                logger.error("The provided tool identifier is invalid.")
-                raise InvalidResourceIdentifier
+                raise InvalidResourceIdentifier(
+                    "The provided tool identifier is invalid"
+                )
             ret_tool_id = match.group('tool_id')
             ret_version_id = match.group('version_id')
 
         if version_id is not None:
             match = re.search(self._RE_VERSION_ID, version_id, re.I)
             if match is None:
-                logger.error("The provided version identifier is invalid.")
-                raise InvalidResourceIdentifier
+                raise InvalidResourceIdentifier(
+                    "The provided version identifier is invalid"
+                )
             ret_version_id = match.group('version_id')
 
         if ret_tool_id is not None:
@@ -504,17 +696,22 @@ class TRSClient():
         self,
         content_accept: str = 'application/json',
         content_type: Optional[str] = None,
+        token: Optional[str] = None,
     ) -> None:
         """Build dictionary of request headers.
 
         Arguments:
             content_accept: Requested MIME/content type.
             content_type: Type of content sent with the request.
+            token: Bearer token for authentication. Set if required by TRS
+                implementation and if not provided when instatiating client or
+                if expired.
         """
         self.headers['Accept'] = content_accept
         if content_type:
             self.headers['Content-Type'] = content_type
-        if self.token:
+        if token is not None:
+            self.token = token
             self.headers['Authorization'] = f"Bearer {self.token}"
 
     def _validate_content_type(
@@ -535,7 +732,92 @@ class TRSClient():
                 content type.
         """
         if requested_type not in available_types:
-            logger.error(
-                "Requested content type not provided by the service."
+            raise ContentTypeUnavailable(
+                "Requested content type not provided by the service"
             )
-            raise ContentTypeUnavailable
+
+    def _send_request_and_validate_response(
+        self,
+        url: str,
+        validation_class_200: Optional[
+            Union[ModelMetaclass, Tuple[ModelMetaclass]]
+        ] = None,
+        validation_class_error: ModelMetaclass = Error,
+        method: str = 'get',
+        payload: Optional[Dict] = None,
+    ) -> Union[str, ModelMetaclass, List[ModelMetaclass]]:
+        """Send a HTTP equest, validate the response and handle potential
+        exceptions.
+
+        Arguments:
+            url: The URL to send the request to.
+            validation_class_200: Type/class to be used to validate a 200
+                response. Either a Pydantic model, a tuple with a Pydantic
+                model as the only item (for list responses), or `None` (for
+                string responses).
+            validation_class_error: Pydantic model to be used to validate
+                non-200 responses.
+            method: HTTP method to use for the request.
+
+        Returns:
+            Unmarshalled response.
+        """
+        # Process parameters
+        validation_type = "model"
+        if isinstance(validation_class_200, tuple):
+            validation_class_200 = validation_class_200[0]
+            validation_type = "list"
+        elif validation_class_200 is None:
+            validation_type = "str"
+        try:
+            request_func = eval('.'.join(['requests', method]))
+        except AttributeError as e:
+            raise AttributeError("Illegal HTTP method provided.") from e
+
+        # Compile request arguments
+        kwargs = {
+            'url': url,
+            'headers': self.headers,
+        }
+        if payload is not None:
+            kwargs['json'] = payload
+
+        # Send request and manage response
+        try:
+            response = request_func(**kwargs)
+        except (
+            requests.exceptions.ConnectionError,
+            socket.gaierror,
+            urllib3.exceptions.NewConnectionError,
+        ):
+            raise requests.exceptions.ConnectionError(
+                "Could not connect to API endpoint"
+            )
+        if not response.status_code == 200:
+            try:
+                logger.warning("Received error response")
+                return validation_class_error(**response.json())
+            except (
+                json.decoder.JSONDecodeError,
+                pydantic.ValidationError,
+            ):
+                raise InvalidResponseError(
+                    "Response could not be validated against API schema"
+                )
+        else:
+            try:
+                if validation_type == "list":
+                    return [
+                        validation_class_200(**obj) for obj in response.json()
+                    ]
+                elif validation_type == "str":
+                    return str(response.json())
+                else:
+                    return validation_class_200(**response.json())
+            except (
+                json.decoder.JSONDecodeError,
+                pydantic.ValidationError,
+            ):
+                raise InvalidResponseError(
+                    "Response could not be validated against API schema"
+                )
